class SpconfigsController < ApplicationController
  require 'utility'
  require 'open-uri'
  require 'nokogiri'
  require 'openssl'

  before_action :set_remote_user, :get_total_number_of_configs
  before_action :set_paper_trail_whodunnit
  before_action :set_is_admin, :set_page_title, :set_breadcrumbs
  before_action :set_spconfig, only: [:show, :edit, :update, :destroy]
#  before_action :set_default_federation
  after_action  :update_partners_list, only: [:create, :update, :destroy]


  # GET /spconfigs
  # GET /spconfigs.json
  # Show only a few buttons: create, search, show-all.
  def index
    # Display the "List ALL Configurations" only when user is some kind of
    # admin.
    @show_all_button = (@is_admin or @is_readonly_admin)
  end

  # GET /spconfigs/list
  # Show everything that the user can see. If there is search parameter, use
  # that to limit the results.
  def list
    username = Utility.get_REMOTE_USER()

    # Only show objects owned by REMOTE_USER or objects with a group
    # that REMOTE_USER is in. However, we allow
    # admins to see all the entries if @is_admin is true.
    if (params[:search]) then
      search_string = params[:search]
      @spconfigs = Spconfig.where("entityid LIKE ?",
                                  "%#{search_string}%")
      @breadcrumbs.push(
        ["All SP Configuration(s) (search string &quot;#{search_string}&quot;)".html_safe, ""]
      )
      flash.notice = "Search found #{@spconfigs.size} Service Provider Configuration(s)"
    else
      user       = Utility.get_REMOTE_USER()
      @spconfigs = Spconfig.this_users_spconfigs(user)
      @breadcrumbs.push(
        ["SP Configuration(s) for #{username}", ""]
      )
      flash.notice = "#{@spconfigs.size} Service Provider Configuration(s)"
    end
  end

  # GET /spconfigs/all
  # Only for admins.
  def all
    username = Utility.get_REMOTE_USER()

    if (@is_admin or @is_readonly_admin) then
      @spconfigs = Spconfig.all()
      @breadcrumbs.push(
        ["All SP Configuration(s) (admin)", ""]
      )
      flash.notice = "#{@spconfigs.size} Service Provider Configuration(s)"
      render 'list'
    else
      redirect_to "/spconfigs/list"
    end
  end

  # GET /spconfigs/1
  # GET /spconfigs/1.json
  def show
    @group_string = @spconfig.workgroup_names.join(', ')

    if (@group_string == '') then
      @group_string = "no owning workgroups"
    end

    # Get the certificates
    @certificates = @spconfig.certificates

    breadcrumb = %Q(SP Configuration <span class="sphighlight">#{@spconfig.entityid}</span>)
    @breadcrumbs.push(
      [breadcrumb.html_safe, ""]
    )
  end

  # GET /spconfigs/new
  def new
    @spconfig = Spconfig.new
    Utility.logger.info(params.inspect)
    @breadcrumbs.push(
      ["New Service Provider Configuration", ""]
    )
  end

  # GET /spconfigs/1/edit
  def edit
    breadcrumb = get_breadcrumb('edit')
    @breadcrumbs.push(
      [breadcrumb, ""]
    )

    if (@spconfig.federation)
      @is_federated = true
    else
      @is_federated = false
    end

  end

  # POST /spconfigs
  # POST /spconfigs.json
  def create
    Utility.logger.info("params is #{params.inspect}")
    Utility.logger.info("spconfig_params is #{spconfig_params.inspect}")

    @spconfig = Spconfig.new(spconfig_params)
    @spconfig.owned_by   = Utility.get_REMOTE_USER_or_autoadmin()
    @spconfig.created_by = Utility.get_REMOTE_USER_or_autoadmin()
    @spconfig.updated_by = Utility.get_REMOTE_USER_or_autoadmin()

    # For reasons beyond my understanding, the owning groups attributes
    # from the submitted web form are not avaialable in the model for
    # validation. But we need them to do proper validation, so to get
    # around this we pass them into the model using
    # 'add_groups_attributes', a method we wrote for this purpose.
    @spconfig.add_groups_attributes(spconfig_params['groups_attributes'])

    if (params['is_federated'] == 'YES') then
      @is_federated = true
      Utility.logger.info("Setting to default federation")
      default_federation = get_default_federation()
      default_federation.spconfigs << @spconfig
    else
      @is_federated = false
      Utility.logger.info("No federation to set")
    end

    respond_to do |format|
      if @spconfig.save
        format.html { redirect_to @spconfig, notice: 'successfully created', alert: '(SP should be available within 15 minutes)' }
        format.json { render :show, status: :created, location: @spconfig }
      else
        format.html { render :new }
        format.json { render json: @spconfig.errors, status: :unprocessable_entity }
      end
    end
  end

  # PATCH/PUT /spconfigs/1
  # PATCH/PUT /spconfigs/1.json
  def update
    @contact_email_old   = @spconfig.contact_email
    @spconfig.attributes = spconfig_params
    @contact_email_new   = @spconfig.contact_email

    @spconfig.updated_by = Utility.get_REMOTE_USER()

    # For reasons beyond my understanding, the owning groups attributes
    # from the submitted web form are not avaialable in the model for
    # validation. But we need them to do proper validation, so to get
    # around this we pass them into the model using
    # 'add_groups_attributes', a method we wrote for this purpose.
    @spconfig.add_groups_attributes(spconfig_params['groups_attributes'])

    if (@spconfig.federation)
      @is_federated = true
    else
      @is_federated = false
    end

    respond_to do |format|
      if @spconfig.save()
        notice = 'SP configuration successfully updated.'
        if (not @is_federated) then
          take_effect = '(Changes should take effect within 15 minutes.)'
        else
          take_effect = ''
        end

        format.html { redirect_to @spconfig, notice: notice, alert: take_effect }
        format.json { render :show, status: :ok, location: @spconfig }
      else
        breadcrumb = get_breadcrumb('edit')
        @breadcrumbs.push(
          [breadcrumb, ""]
        )
        format.html { render :edit }
        format.json { render json: @spconfig.errors, status: :unprocessable_entity }
      end
    end
  end

  # DELETE /spconfigs/1
  # DELETE /spconfigs/1.json
  def destroy
    @contact_email_old = @spconfig.contact_email
    @contact_email_new = nil

    @spconfig.destroy
    respond_to do |format|
      format.html { redirect_to spconfigs_url, notice: 'SP Config was successfully removed.' }
      format.json { head :no_content }
    end
  end

  private
    # Use callbacks to share common setup or constraints between actions.
    def set_spconfig
      @spconfig = Spconfig.find(params[:id])
      if (!can_access()) then
        # We allow the show action, but not any other actions
        if (action_name != 'show') then
          redirect_to "/errors/forbidden"
          return
        end
      end
    end

    def set_page_title
      @page_title      = Rails.application.config.page_title
      @page_title_path = spconfigs_path
    end

    # Does the current user (i.e., SUNetID) "own" this configuration?
    # That is, does the current user satisfy one (or more) of the
    # following:
    #
    #   1. user matches owned_by
    #   2. user is an ADMIN
    #   3. user is in one of the owning groups
    #
    # If yes, return true, otherwise return false.
    def can_access()
      user = Utility.get_REMOTE_USER()

      # 1. user is the owner
      if (@spconfig.owned_by == user) then
        return true
      end

      # 2. user is an SPDB admin.
      if (Utility.is_admin(user)) then
        return true
      end

      # 3. user is one (or more) of this models groups.
      if (@spconfig.user_in_groups(user)) then
        return true
      end

      return false
    end

    def update_partners_list()
      # If the action is update or create, add to the partners list.
      if (action_name == 'create') then
        # CREATE
        Utility.add_email_to_partners_mailman_list(@spconfig.contact_email, nil)
      elsif (action_name == 'update') then
        # UPDATE
        # Did the contact_email attribute change?
        if (@contact_email_old != @contact_email_new) then
          Utility.logger.info("contact_email changed: #{@contact_email_old} (old) --> #{@contact_email_new} (new) ")
          Utility.add_email_to_partners_mailman_list(@contact_email_new, @contact_email_old)
        else
          Utility.logger.info("contact_email did NOT change")
        end
      elsif (action_name == 'destroy') then
        # DESTROY
        Utility.logger.info("contact_email in destryed spconfig is #{@contact_email_old}")
        Utility.remove_unused_email_from_partners_list(@contact_email_old)
      end

      return
    end

    def get_total_number_of_configs()
      @total_number_of_configs = Spconfig.count()
    end

    # Never trust parameters from the scary internet, only allow the white list through.
    def spconfig_params
      params.require(:spconfig).permit(:entityid, :metadata, :owned_by,
                                       :contact_email, :login_url,
                                       :comments,
                                       :sign_responses,
                                       :sign_assertions,
                                       :encrypt_assertions,
                                       :is_legacy,
                                       :is_federated,
                                       :ignore_cert_issues,
                                       :ignore_valid_until_issues,
                                       groups_attributes: [:id, :workgroup, :_destroy],
                                      )
    end

    def set_breadcrumbs
      @breadcrumbs = Utility.first_level_breadcrumb
      @breadcrumbs.push(['SP Configurations', '/spconfigs/index'])
      @model_name = 'spconfig'
    end

    def get_default_federation
      @default_federation = Federation.find_by name: 'InCommon'
    end

    def get_breadcrumb(action)
      if (action == 'edit') then
        breadcrumb = %Q(Edit SP <span class="sphighlight">#{@spconfig.entityid}</span>)
        return breadcrumb.html_safe
      end
    end

end
