require 'net/http'
require 'net/https'
require 'uri'
require 'json'
require 'utility'
require 'digest'

class Spconfig < ActiveRecord::Base
  # Each spconfig can zero, one, or many owning groups, but we don't want
  # duplicates.
  has_many :groups

  belongs_to :federation

  # Each spconfig can zero, one, or many certificates. We want those
  # certificates to go away when the spconfig goes away.
  has_many :certificates, dependent: :destroy

  accepts_nested_attributes_for :groups, reject_if: :all_blank, allow_destroy: true

  strip_attributes

  # We want a place to put the groups_attributes passed in from the
  # controller.
  attr_accessor :xgroups_attributes

  # We want to give admins the (hopefully rarely used) option to supress
  # the validity check on certificate expiration, existence, etc.
  attr_accessor :ignore_cert_issues
  attr_accessor :ignore_valid_until_issues

  # We keep track of versions with paper_trail
  has_paper_trail

  ## VALIDATIONS
  validate  :metadata_is_valid
  validate  :at_least_one_owning_group
  validates :contact_email, :presence => {message: 'You must supply a valid mailing list address'}
  validates :entityid, uniqueness: {message: 'This entityID is already being used by another configuration'},
            if: 'entityid.present?'

  ## CALLBACKS
  before_validation :get_validation_results

  before_destroy do
    self.versions.destroy_all
  end

  # We want to mark for export the metadata files after any create or delete.
  after_create  :after_create_callback
  after_destroy :after_destroy_callback

  # We want to strip out any Signatures before saving.
  before_save   :before_save_callback

  # We want to mark for export the metadata file after any save where there is
  # a metadata change.
  after_save    :after_save_callback

  # A few useful instance variables
  @metadata_validation_results = nil
  @sp_info                     = nil
  @saml_validation_errors      = nil

  # Here is where we define the mapping from metadata type to the basename
  # of its output file.
  @@metadata_type_to_basename =  {
      'combined'   => 'metadata',
      'legacy'     => 'metadata-legacy',
      'non-legacy' => 'metadata-blanket',
  }
  # Each metadata file has a "Name" parameter in the header and this name
  # must be unique for each metdata file consumed by the IdP. Thus, we
  # define a unique header name for each metadata file we output.
  @@metadata_type_to_headername =  {
      'combined'   => 'urn:mace:stanford.edu:shibboleth:stanford',
      'legacy'     => 'https://spdb.stanford.edu/spmetadata/metadata-legacy.xml',
      'non-legacy' => 'https://spdb.stanford.edu/spmetadata/metadata-blanket.xml',
  }

  # Only call this when metadata changes
  def metadata_is_valid
    # If the entityID is not defined, stop since other validation tests depend on
    # there being an entityID.
    if (not self.entityid) then
      errors.add(:base, "Missing entityId.")
    end

    # Deal with federated entries first.
    if (self.federation) then
      # Check that this entityId is in the incommon metdata. If not, add
      # error.
      if (not Utility.is_an_incommon_entity_id(self.entityid)) then
        errors.add(:base, "This entityId is not in my #{self.federation.name} metadata.")
      end

      Utility.logger.info("has an associated federation so skipping metadata validation")
      # Since there is an associated federation, we skip all metadata validation.
      self.metadata = "SP metadata provided by the #{self.federation.name} federation"
      return
    else
      Utility.logger.info("does not have an associated federation")
    end

    # Skip validation of the metadata if this is an old record AND the
    # metadata field has NOT changed.
    if (!self.metadata_was.nil? and !self.metadata_changed?) then
      return
    end

    # If the user has indicated that this SP is part of the InCommon Federation,
    # set metatadata to some default string and skip the rest of the validation.
    if (metadata.nil? or metadata.empty?) then
      errors.add(:base, "Metadata cannot be empty.")
      return
    end

    # If the whitespace-trimmed version of the metadata from BEFORE the
    # edit agrees with the whitespace-trimmed version of the metadata from
    # AFTER the edit, then there is no reason to validate
    if (self.metadata_was and
        (Utility.normalize_whitespace(self.metadata_was) == Utility.normalize_whitespace(metadata))) then
      return
    end

    length = @metadata_validation_results.length()

    if (length != 2) then
      raise 'metadata validator failed'
    end

    # Set the ignore_certificate_expiration and ignore_no_certificates
    # booleans based on the ignore certificate issues check box. This
    # checkbox only appears for admins.
    if ((not Rails.configuration.check_cert_expirations) or
        (self.ignore_cert_issues.to_i() == 1)) then
      ignore_certificate_expiration = true
    else
      ignore_certificate_expiration = false
    end

    if ((not Rails.configuration.check_for_encrypting_cert) or
        (self.ignore_cert_issues.to_i() == 1)) then
      ignore_no_encrypting_certificates = true
    else
      ignore_no_encrypting_certificates = false
    end

    # Set the ignore_valid_until_expiration boolean based on the
    # global Rails configuration and checkbox.
    if ((not Rails.configuration.check_valid_until_expirations) or
        (self.ignore_valid_until_issues.to_i() == 1)) then
      ignore_valid_until_expiration = true
    else
      ignore_valid_until_expiration = false
    end


    # Utility.logger.info(@sp_info.inspect())
    # Utility.logger.info(@saml_validation_errors.inspect())

    # Go through the validation errors adding them as appropriate.
    if (!@saml_validation_errors.nil?)
      if (@saml_validation_errors.key?('200')) then
        # If the metadata contains '<?xml' skip the other errors.
        errors.add(:base, "Metadata must NOT contain the string '<?xml'")
        return
      elsif (@saml_validation_errors.key?('300')) then
        # If the metadata is not valid XML skip the other errors.
        errors.add(:base, 'Metadata is not valid XML')
        return
      elsif (@saml_validation_errors.key?('350')) then
        # If the metadata is not valid SAML XML skip the other errors.
        error_value = @saml_validation_errors['350'][0]
        error_value = error_value.strip()

        # Remove everything before the first colon (this gets rid of the
        # extra "Metadata is not valid SAML 2.0 XML".
        error_value.gsub!(/^[^:]*:/, '')

        errors.add(:base, "350: #{error_value}")
        return
      else
        # For the other errors, add them now. Note that we skip certificate
        # errors (5xx) and deal with them below.
        # Sort the errors.
        error_numbers = @saml_validation_errors.keys.sort
        error_numbers.each do |error_number|
          if (error_number =~ /^5/) then
            # This is a certificate error, so skip (for now).
            next
          end
          if (error_number == '410') then
            # This is a validUntil error we which we handle ourselves
            # below, so skip (for now).
            next
          end
          error_texts = @saml_validation_errors[error_number]
          error_texts.each do |error_text|
            errors.add(:base, "Metadata [#{error_number}]#{error_text}")
          end
        end
      end
    end

    ## ## INCOMMON METADATA
    # Don't allow submission of metadata using an InCommon entityID (may
    # have to change this later).
    if (Utility.is_an_incommon_entity_id(self.entityid)) then
      errors.add(:base, "The entityID supplied ('#{self.entityid}') " +
                        "is an InCommon entityID; this SP's metadata is already known to our IdPs.")
    end

    ## ## ENCRYPTING CERTIFICATE
    if (not ignore_no_encrypting_certificates) then
      # Get the certificates (if any)
      certificates = Utility.get_certificate_objects_from_metadata(metadata)

      # Count number of certificate uses.
      num_certificates_encryption  = 0
      num_certificates_signing     = 0
      num_certificates_unspecified = 0
      certificates.each do |certificate|
        use = certificate.use
        if (certificate.use_is_encryption?())
          num_certificates_encryption = num_certificates_encryption + 1
        elsif (certificate.use_is_signing?())
          num_certificates_signing = num_certificates_signing + 1
        elsif (certificate.use_is_unspecified?())
          num_certificates_unspecified = num_certificates_unspecified + 1
        else
          errors.add(:base, "One of your certificates has an unrecognized use")
        end
      end

      Utility.logger.info "\# certificates (encryption):  #{num_certificates_encryption}"
      Utility.logger.info "\# certificates (signing):     #{num_certificates_signing}"
      Utility.logger.info "\# certificates (unspecified): #{num_certificates_unspecified}"

      if (certificates.count == 0) then
        errors.add(:base, "Metadata must contain at least one certificate for encryption")
      elsif ((num_certificates_encryption == 0) and (num_certificates_unspecified == 0))
        errors.add(:base, "Metadata must contain at least one certificate for encryption")
      else
      end

    end

    # These are used in a couple of the following sections that check for
    # various kinds of expirations. Because some SP's generate a
    # certificate for exactly one year, we can get submissions that are a
    # few hours short of one year. So, set our limit to be one year minus
    # 14 days.
    almost_one_year_of_seconds = 60 * 60 * 24 * (365 - 14)
    now_utc                    = Time.now.utc
    expiration_limit_utc       = now_utc.to_i() + almost_one_year_of_seconds

    ## ## CERTIFICATE EXPIRATION
    if (not ignore_certificate_expiration) then

      # This is a hash mapping use type to the number of certificates
      # of that type that are not expired. Note that if no certificates
      # of that use type appear, there will be no key for that use type.
      cert_use_count = {}

      # Check certificate expirations. For each KeyDescriptor, at least one
      # certificate must be valid for a year.
      if (@sp_info.key?('KeyDescriptors')) then
        key_descriptors = @sp_info['KeyDescriptors']

        key_descriptors.each do |key_descriptor|
          usage = nil
          if (key_descriptor.key?('usage')) then
            usage = key_descriptor['usage']
          end

          if (usage.nil?() or usage.strip().empty?()) then
            usage = 'not specified'
          end

          # Look at all the certificates
          if (key_descriptor.key?('X509Certificates')) then
            x509Certificates = key_descriptor['X509Certificates']

            x509Certificates.each do |x509Certificate|
              if (!cert_use_count.key?(usage)) then
                # This is the first time we have encountered a certificate
                # of type usage.
                cert_use_count[usage] = 0
              end

              notAfter_epoch = x509Certificate['notAfter_epoch']

              if (ignore_certificate_expiration or (expiration_limit_utc < notAfter_epoch)) then
                # This certificate is OK, so we increment the counter that keeps track of
                # how many certificates of this usage type are still withing the expiration
                # window.
                cert_use_count[usage] = cert_use_count[usage] + 1
              else
                # This certificate is expired.
              end
            end
          end
        end
      end

      # Loop through all the usage types we might have found.
      cert_use_count.keys.each do |usage_type|
        if (cert_use_count[usage_type] == 0) then
          errors.add(:base,
                     "No certificate found with at least one year left before expiration (use: #{usage_type})")
        end
      end
    end

    ## ## VALIDUNTIL ATTRIBUTE
    # Check the validUntil attribute.
    # If the validUntil attribute exists and is less than our expiration limit raise a
    # validation error.
    if (not ignore_valid_until_expiration) then
      if (not self.valid_until.blank?()) then
        valid_until_expiration_limit_utc = now_utc.to_i() + Rails.configuration.check_valid_until_expired_seconds
        if (self.valid_until.to_i() < valid_until_expiration_limit_utc) then
              expired_text = Rails.configuration.check_valid_until_expired_text
              errors.add(:base,
                         "Metadata's validUntil attribute (#{@sp_info['validUntil']}) does not have at least #{expired_text} left before expiration")
        end
      end
    end

    ## ## ACSF ENTITYID
    # Make sure this entityId does not match the format used by ACSF.
    if (self.entityid =~ /^https:\/\/acsf.stanford.edu\/saml\/.*$/) then
      errors.add(:base, "Your entityId '#{self.entityid}' is reserved for Acquia Cloud Sites (ACSF)")
    end

  end


  # Function to add a parameter (used to pass information from controller
  # to model).
  def add_groups_attributes(groups_attributes)
    if (not groups_attributes.nil?) then
      self.xgroups_attributes = groups_attributes
    else
      self.xgroups_attributes = {}
    end
    return
  end

  def at_least_one_owning_group()
    groups_attributes = self.xgroups_attributes
    Utility.logger.info "xgroups_attributes is #{xgroups_attributes.inspect}"
    groups_attributes.each do |key, group_attr|
      # group_attr is a hash. At least one must have its key
      # '_destroy' not be 1.
      if (group_attr.has_key?('_destroy')) then
        destroy_value = group_attr['_destroy']
        if (destroy_value != "1") then
          # At least one is not to be destroyed, so we are
          # good.s
          return
        end
      end
    end

    # If we get here, either there were no owning groups, or else there
    # were owning groups but they were all to be destroyed. In either
    # case, that is a problem.
    errors.add(:base, "Must have at least one owning group")
    return
  end

  #def valid_until()
  #  if (@sp_info.nil?()) then
  #    self.get_validation_results()
  #  end

  #  if (@sp_info.key?('validUntil')) then
  #    return @sp_info['validUntil']
  #  else
  #    return nil
  #  end
  #end

  def to_s()

    print "entityid:           #{self.entityid}\n"
    print "contact_email:      #{self.contact_email}\n"
    print "owned_by:           #{self.owned_by}\n"
    print "loginURL:           #{self.login_url}\n"
    print "created_at:         #{self.created_at}\n"
    print "updated_at:         #{self.updated_at}\n"
    print "created_by:         #{self.created_by}\n"
    print "updated_by:         #{self.updated_by}\n"
    print "sign_responses:     #{self.sign_responses}\n"
    print "sign_assertions:    #{self.sign_assertions}\n"
    print "encrypt_assertions: #{self.encrypt_assertions}\n"
  end


  # Write the metadata file (maybe). Return true if the file was written,
  # false otherwise.
  def self.write_metadata_file(metadata_type, force_update = false)

    # First of all, do we even _need_ to write the metadata file?
    if (not Spconfig.should_metadata_file_be_exported(metadata_type, force_update)) then
      Utility.logger.info "no need to update metadata_type '#{metadata_type}' at this time"
      return false
    else
      Utility.logger.info "metadata_type '#{metadata_type}' needs to be updated"
    end

    # If we get here then we need to write the file.

    # Step 1. Define the path where the output file will go.
    filepath = self.metadata_type_to_filepath(metadata_type)

    # f will be a temporary file for the complete metadata file.
    f = Tempfile.new(metadata_type)

    # Step 2. Create a raw file of all the metadata. This will vary depending
    # on the metadata_type.
    raw_metadata_file = Tempfile.new(metadata_type)
    raw_metadata_file.close()
    Spconfig.create_raw_metadata_file(metadata_type, raw_metadata_file)


    # Step 3. Construct the complete output file.

    # Step 3a. Make the header.
    # HEADER
    # We need to replace the string '%%VALID_UNTIL%%' with a valid until
    # time in form "2016-10-10T18:43:47Z". How long this file is valid is
    # set as a number of seconds in the application.rb configuration file.
    header = Rails.configuration.sp_metadatafile_header.dup

    expiration_time    = Time.now.utc + Rails.configuration.sp_metadatafile_expire_seconds
    valid_until_string = expiration_time.strftime('%FT%TZ')

    header.sub!('%%VALID_UNTIL%%', valid_until_string)

    # We also need to replace the string '%%NAME%%' with the correct
    # headername (see the definition of @@metadata_type_to_headername
    # above for more details).
    headername = @@metadata_type_to_headername[metadata_type]
    header.sub!('%%NAME%%', headername)

    # We also need to replace "%%ID_REPLACE%%" with a unique
    # identifier. We use a timestamp not for any technical reason but just
    # to make it easier to determine when the metadata file was created.
    # Example:
    # ID="20161014075412MDrrr". The "MD" is the literal string "MD" and
    # is there to indicate this is MetadData. The "rrr" are three random
    # letters to make extra sure there is no ID name collision.
    current_time = Time.now.utc
    rrr          = (0...3).map { ('a'..'z').to_a[rand(26)] }.join
    id_string    = current_time.strftime('%Y%m%d%H%M%S') + 'MD' + rrr

    # Now do the replace.
    header.sub!('%%ID_REPLACE%%', id_string)

    f.write(header)
    f.write("\n")

    # Step 3b. Put in all the actual metadata.
    # MIDDLE (the raw metadata strings)
    File.open(raw_metadata_file, 'r') do |rf|
      rf.each_line do |line|
        f.write(line)
      end
    end

    # Step 3c. Put in the footer.
    # FOOTER
    f.write("\n")
    f.write(Rails.configuration.sp_metadatafile_footer)

    f.close()

    Utility.logger.info "wrote complete unsigned metadata file to '#{f.path()}'"

    # Step 4. Remove the raw metadata file.
    # At this point, we can delete the temporary file raw_metadata
    raw_metadata_file.unlink()

    # Step 5. Sign the output file (maybe) and move the file into place.
    if (Rails.configuration.sp_metadatafile_sign) then
      stdout, stderr, status = Utility.sign_xml_file(f.path(), filepath)
      if (status.exitstatus != 0) then
        Utility.logger.info "signing file error: [stdout: #{stdout}][stderr: #{stderr}][status: #{status}]"
      else
        Utility.logger.info "signed metadata file and stored in '#{filepath}'"
      end
    else
      File.rename(f.path(), filepath)
      Utility.logger.info "not signing metadata file; moved temporary unsigned metadata file to '#{filepath}'"
    end

    # Cleanup the temporary file
    if (File.exist?(f.path())) then
      f.unlink()
    end

    # We return true since the file was written.
    return true
  end

  def self.metadata_type_to_filepath(metadata_type)
    basepath = Rails.configuration.sp_metadatafile_basedir

    # Get the basename for this metadata type.
    basename = @@metadata_type_to_basename[metadata_type]

    basename_xml = "#{basename}.xml"
    filepath     = File.join(basepath, basename_xml)

    return filepath
  end


#  # Return true if the specified metadata file is due to be exported.
#  #
#  # The metadata_type parameter must be one of these values:
#  #    * 'combined'   - all the metadata
#  #    * 'legacy'     - all the metadata whose is_legacy field is TRUE
#  #    * 'non-legacy' - all the metadata whose is_legacy field is FALSE
#  #
#  # Here is the logic.
#  #  IF force_update is true
#  #    THEN the target metadata file will be created/updated
#  #  ELSE
#  #    IF the target metadata file DOES NOT exist
#  #      THEN target metadata file will be created.
#  #    ELSE
#  #      IF the last updated time in the updatetime table is NEWER than
#  #      the target metadata file
#  #        THEN the target metadata file will be updated
#  #      ELSE
#  #        IF the target metadata file is older than 2 hours
#  #          THEN the target metadata file will be updated
#  #        ELSE
#  #          The target metadata file will NOT be updated.
#  #
#  def self.should_metadata_file_be_exported(metadata_type,
#                                       force_update     = false,
#                                       updatedtime_name = 'last-changed'
#                                      )
#
#    filepath = Spconfig.metadata_type_to_filepath(metadata_type)
#
#    write_new_metadata = nil
#
#    if (force_update) then
#      # force_update is true, then write metadata.
#      write_new_metadata = true
#      Utility.logger.info "force_update is true, so will create/update"
#    else
#      if (not File.exist?(filepath)) then
#        # No existing file? Then write metadata.
#        write_new_metadata = true
#        Utility.logger.info "metadata file does not exist, so will create"
#      else
#        # The file exists. Is it older than the time in the updatedtime
#        # table?
#        last_updated_time = Updatedtime.last_updated()
#        if (last_updated_time.nil?() or (File.mtime(filepath) < last_updated_time)) then
#          # The existing file is older, so update.
#          write_new_metadata = true
#          Utility.logger.info "metadata file is older than last-update time, so will update"
#        else
#          # The file exists but is not older than the last updated time. How old is it?
#          file_age_seconds = (Time.now - File.mtime(filepath)).to_i()
#
#          # If older than 2 hours, write metadata.
#          if (file_age_seconds > 60 * 60 * 2) then
#            write_new_metadata = true
#            Utility.logger.info "metadata file exists but is old enough to require an update"
#          else
#            Utility.logger.info "metadata file exists and does not need to be updated"
#          end
#        end
#      end
#    end
#
#    return write_new_metadata
#  end


  # The metadata_type parameter must be one of these values:
  #    * 'combined'   - all the metadata
  #    * 'legacy'     - all the metadata whose is_legacy field is TRUE
  #    * 'non-legacy' - all the metadata whose is_legacy field is FALSE
  #
  def self.should_metadata_file_be_exported(
        metadata_type,
        force_update = false
      )
    metadata_filepath = Spconfig.metadata_type_to_filepath(metadata_type)

    return Utility.should_file_be_exported(
             filepath:           metadata_filepath,
             updatedtime_name:   'last-changed',
             force_update:       force_update,
             time_limit_seconds: 2 * 60 * 60,
             action_label:       'metadata-changes'
           )
  end


  def self.create_raw_metadata_file(metadata_type, filepath)
  # Read the metadata from all the records into a file in a normalized
  # order. Returns the path to the supplied filepath.
  #
  # Remember that we DON'T want to include metadata from an spconfig that is
  # part of a federation.
    if (metadata_type == 'combined') then
      spconfigs = Spconfig.where(federation_id: nil).order(:id).all()
    elsif (metadata_type == 'legacy') then
      spconfigs = Spconfig.where(federation_id: nil, is_legacy: true).order(:id).all()
    elsif (metadata_type == 'non-legacy') then
      spconfigs = Spconfig.where(federation_id: nil, is_legacy: false).order(:id).all()
    else
      raise "unrecognized metadata type #{metadata_type}"
    end

    File.open(filepath, 'w') do |f|
      spconfigs.each do |spc|
        f.write(spc.metadata)
      end
    end

    return filepath
  end

  # Save all the metadata_types to their respective files. We return a hash that maps
  # the metadata_type to a boolean. Mapping to true means the file was actually written,
  # false means the write was skipped.
  def self.save_metadatas_to_file(force_update = false)
    metadata_type_to_was_written = {}
    @@metadata_type_to_basename.each do |metadata_type, basename|
      Utility.logger.info "creating (if necessary) metadata output file for metadata_type '#{metadata_type}'"
      rv = Spconfig.write_metadata_file(metadata_type, force_update)
      metadata_type_to_was_written[metadata_type] = rv
    end
    return metadata_type_to_was_written
  end


  # Return true if user is in at least one of this object's groups
  def user_in_groups(user)
    groups.each do |group|
      Utility.logger.info "checking to see if #{user} is in group #{group.workgroup}"
      if (Utility.user_member_of_privgroup_cached(user, group.workgroup))
        return true
      end
    end

    return false
  end

  # Return an array of the spconfigs that the given user has access
  # to. That is all the spconfigs that the user owns, or that contain a
  # group that the user is a member of.
  def self.this_users_spconfigs(user)
    # Step 1. Get all the spcongifs owned by this user.
    spconfigs_owned = Spconfig.where(owned_by: user)

    # Collect all the id's from @spconfigs_owned so that we can avoid
    # duplicating them in @spconfigs_group_owned.
    spconfigs_owned_ids = {}
    spconfigs_owned.each do |spconfig|
      spconfigs_owned_ids[spconfig.id] = true
    end

    # 2. Now we find all spconfigs that have an associated group
    # containing user as a member.
    spconfigs_group_owned = []

    spconfigs_all = Spconfig.all()
    spconfigs_all.each do |spconfig|
      if (spconfig.user_in_groups(user)) then
        if (not spconfigs_owned_ids.key?(spconfig.id)) then
          spconfigs_group_owned.push(spconfig)
        end
      end
    end

    spconfigs = spconfigs_owned + spconfigs_group_owned
    return spconfigs
  end


  # Return an array of the workgroup names. Will return the empty
  # array if there are none.
  def workgroup_names()
    if (groups.nil?() or groups.empty?())
      return []
    else
      return groups.map{ |g| g.workgroup }
    end
  end

  def after_create_callback
    Spconfig.save_metadatas_to_file()
  end

  def after_destroy_callback
    Spconfig.save_metadatas_to_file()
    Utility.logger.info 'configuration deleted; touching the updatedtimes table'
    Updatedtime.touch()
  end

  # Call this method AFTER any save.
  def after_save_callback
    change_detected = false

    # Did the metadata field change?
    if (not self.metadata_change.nil?) then
      change_detected = true
      Utility.logger.info "change in 'metadata' field detected"
    end

    # Did the is_legacy field change?
    if (not self.is_legacy_change.nil?) then
      change_detected = true
      Utility.logger.info "change in 'is_legacy' field detected"
    end

    if (change_detected) then
      Utility.logger.info 'XML-affecting change detected; touching the updatedtimes table'
      Updatedtime.touch()
    else
      Utility.logger.info 'No XML-affecting change detected; not touching the updatedtimes table'
    end

    # Remove any duplicate workgroups.
    workgroup_hash = {}
    self.groups.each do |group|
      workgroup = group.workgroup
      if (workgroup_hash.has_key?(workgroup)) then
        # This is already in the list, so delete
        group.delete()
        Utility.logger.info("deleting duplicate workgroup '#{workgroup}'")
      else
        workgroup_hash[workgroup] = true
      end
    end

    # Add the certificates (if any)
    certificates = Utility.get_certificate_objects_from_metadata(metadata)

    self.certificates.destroy_all
    self.certificates << certificates

  end

  # Call this method BEFORE any save.
  def before_save_callback
    # 1. Strip any Signature elements
    if (not self.federation) then
      stripped_string = Utility.strip_signatures(metadata)
      if (!stripped_string.nil?()) then
        # At least one Signature removed.
        self.metadata = stripped_string
      end

      # 2. Remove the XML declaration <?xml ...> if there is one.
      self.metadata = try_to_strip_xml_declaration(self.metadata)
    end
end

  protected
  def get_validation_results
    if (self.federation) then
      Utility.logger.info("entityid is #{self.entityid}")
    else
      metadata_stripped = try_to_strip_xml_declaration(metadata)

      @metadata_validation_results = Utility.submit_metadata_for_validation(metadata_stripped)

      @sp_info                = @metadata_validation_results[0]
      @saml_validation_errors = @metadata_validation_results[1]

      Utility.logger.info("metadata validation results: #{@metadata_validation_results.inspect}")
      # Pre-populate entityid so we we can find it.
      if (@sp_info.key?('entityID'))
        self.entityid = @sp_info['entityID']
      end

      # Also populate validUntil (if it exists)
      if (@sp_info.key?('validUntil') and !@sp_info['validUntil'].empty?()) then
        self.valid_until = DateTime.xmlschema(@sp_info['validUntil'])
      end
    end
  end

  # Returns the XML without the <?xml header if the XML is valid. If the
  # XML is _not_ valid simply returns the original XML string.
  def try_to_strip_xml_declaration(metadata_string)
    stripped_metadata = nil
    begin
      stripped_metadata = Utility.strip_xml_declaration(metadata_string)
    rescue
      stripped_metadata = nil
    end

    if (stripped_metadata.nil?()) then
      return metadata_string
    else
      return stripped_metadata
    end
  end


end
